import torch
import nvdiffrast.torch as dr
import numpy as np
import trimesh


def show_mesh(vertices, faces, vtx_colors=None):
    if isinstance(vertices, torch.Tensor):
        vertices = vertices.detach().cpu().numpy()

    if isinstance(faces, torch.Tensor):
        faces = faces.detach().cpu().numpy()

    if isinstance(vtx_colors, torch.Tensor):
        vtx_colors = vtx_colors.detach().cpu().numpy()

    mesh = trimesh.Trimesh(vertices=vertices, faces=faces, vertex_colors=vtx_colors)
    trimesh.Scene([mesh]).show()


def get_invisible_faces(rast_tri_ids, faces):
    #tri_ids = torch.unique(rast_tri_ids.flatten().to(torch.long))
    visibles = torch.unique(rast_tri_ids.flatten().to(torch.long))
    invisibles = sorted(list(set(range(len(faces))) - set(visibles.tolist())))
    invisibles = torch.tensor(invisibles, dtype=torch.long).to(visibles.device)

    return visibles, invisibles


def main():
    ndc_data = np.load('ndc.npy', allow_pickle=True)

    vtx_pos = ndc_data.item()['vtx_pos']
    faces = ndc_data.item()['faces']
    vtx_colors = ndc_data.item()['vtx_color']
    #show_mesh(vtx_pos[0, :, :3], faces, vtx_colors)
    
    use_opengl = False
    glctx = dr.RasterizeGLContext() if use_opengl else dr.RasterizeCudaContext()

    vtx_pos = torch.tensor(vtx_pos).cuda()
    faces = torch.tensor(faces).cuda()
    vtx_colors = torch.tensor(vtx_colors).cuda()

    # not rasterize the 2000th face
    #rast_out, _ = dr.rasterize(glctx, vtx_pos, faces, resolution=(2048, 2048))
    rast_out, _ = dr.rasterize(glctx, vtx_pos, faces[:1999], resolution=(2048, 2048))
    tri_ids = rast_out[..., 3]
    visibles, invisibles = get_invisible_faces(tri_ids, faces)
    print("full rasterization:")
    print("visible faces:{}, invisible faces: {}".format(len(visibles), len(invisibles)))

    #show_mesh(vtx_pos[0, :, :3], faces[visibles], vtx_colors)
    #show_mesh(vtx_pos[0, :, :3], faces[invisibles], vtx_colors)
    #assert len(visibles) == 1999, "not all faces rasterized"

    # rasterize invisible face
    rast_out, _ = dr.rasterize(glctx, vtx_pos, faces[invisibles], resolution=(2048, 2048))
    tri_ids = torch.unique(rast_out[..., 3].to(torch.long))
    print("rasterize invisible faces again:")
    print("visible faces:{}, invisible faces: {}".format(len(tri_ids), len(invisibles) - len(tri_ids)))
    if len(tri_ids) != 0:
        print("error for invisible faces appeared now!!!")
    #assert len(tri_ids) == 0, "some invisible face appear now!!!"


if __name__ == "__main__":
    main()
